// lfstream.h
// (c) Petr Mikulik (mikulik@physics.muni.cz) - 1994/1996/1999
// Freeware version. It is distributed without any warranty.

#ifndef LFSTREAM_H
#define LFSTREAM_H

#include <stdio.h>
#include <fstream.h>
#include <iostream.h>
#include "ansicpp.h"


#if defined(__GNUC__) && __GNUC__ >= 3
#define GNU_LIBSTD_3 1
#endif


/*
Usage and comments: use this class for input from a file with such a syntax:

1.54  this is wavelength
! this line will be skipped because double is required
10.0  this is inclination angle
# please, now enter file name
out.dat  usually out.dat

In you program, use
lfstream inp;
double WaveLength, Inclination;
inp >> WaveLength; inp.ligno();  // skip rest of the line + increment line number
inp >> Inclination; inp.ligno(); //   "    and the comment line is autodetected

etc.
If there is an error: the correct line number will be outputed
if there were not empty lines used as white spaces.

To avoid skipping chars and strings which are preceded
by LFSTREAM_LINE_SKIP character, use something like
inp.ifstream >> myString;
*/


#if 1
  #define LFSTREAM_SKIP_LINE "#!%"
#else
  #define LFSTREAM_SKIP_LINE "#!|%"
#endif
// these are the chars denoting a comment line

extern char lfstream_space_after_number /* = 0 */;


class lfstream: public ifstream {
  public:
    unsigned long int line; // line number
    char space_after_number;
	// 1: will require isspace() character after a number
	// 0: can read numbers separated by commas or whatever else
	// Default value in each lfstream is given by the global variable
	// lfstream_space_after_number
    char can_exit;
	// by default, it is set to 1 so that exit(i) is allowed
    int last_error;

  private:
    FILE *inputPipe; //used if the input is pipe or .gz file

  public:
    // overloaded routines:
    lfstream (): ifstream()
      { inputPipe = NULL; last_error=0; can_exit=1; }
#ifndef GNU_LIBSTD_3 
    lfstream (int file_no): ifstream(file_no)
      { inputPipe = NULL; line=1; space_after_number=0;
	last_error=0; can_exit=1;
	tellg();
      }
#endif
#ifndef GNU_LIBSTD_3 
    // Worked alway, but now not for g++ >= 3.0
    lfstream (const char* c, int i = ios::in | ios_bin, int j = filebuf::openprot)
      : ifstream(c,i,j)
#else
    lfstream (const char* c, int i = ios::in | ios_bin )
      : ifstream(c, (ios_base::openmode)i)
#endif
      { inputPipe = NULL; if (fail()) CannotOpen(c);
	line = 1; space_after_number = lfstream_space_after_number;
	last_error = 0; can_exit = 1;
	tellg();
      }

    void open ( char *InputFile );
      //opens an ordinary file, or file.gz, or stdin, or pipe from '<'
    void open_normalfile (const char* c)
      { ifstream::open(c,ios::in|ios_bin); //open in binary, so that CRLF pass ok
	if (fail()) CannotOpen(c);
	inputPipe = NULL; line=1;
	space_after_number=lfstream_space_after_number;
	last_error=0; can_exit=1;
	tellg();
      }
    void attach (int fileDesc, const char *InputFile = 0)
      { 
#ifdef GNU_LIBSTD_3 
	cerr <<"FATAL: file descriptors not supported!\n";
	exit(1);
#else
	ifstream::attach(fileDesc);
#endif
	if (fail()) { CannotOpen( InputFile ? InputFile : "" ); }
	inputPipe = NULL; line=1;
	space_after_number=lfstream_space_after_number;
	last_error=0; can_exit=1;
	tellg();
      }

#ifndef GNU_LIBSTD_3 
    // Worked alway, but now not for g++ >= 3.0
    void open_normalfile (const char* c, int i, int j=filebuf::openprot)
      { ifstream::open(c,i,j);
#else
    void open_normalfile (const char* c, int i )
      { ifstream::open(c,(ios_base::openmode)i);
#endif
	if (fail()) CannotOpen(c);
	inputPipe = NULL; line = 1;
	space_after_number = lfstream_space_after_number;
	last_error = 0; can_exit = 1;
	tellg();
      }
    //note: tellg(); is a trick to make gcc/emx read both LF and CRLF text files

    void close ()
      { /* if (inputPipe!=NULL) pclose(inputPipe); //No! */
	inputPipe = NULL;
	ifstream::close();
      }

    // new routines:
    void ligno ();       // test no param+ignore line+increment line number
    void xligno ();      // ignore line+increment line number
    void errorline ();
    void errorline ( const char *what );
    void error ( const char *what );
    void print_Error_on_line ( const char *what1 = 0, const char *what2 = 0, const char *what3 = 0 );
    void unknown ( const char *what );
    void unknown ( const char *what1, const char *what2 );
    void required ( const char *what );
    void labelrequired ( const char *what );
    void labelrequired ( const char *what, const char *found );
    void Wrong ( const char *what );
    void Wrong ( const char *what1, const char *what2 );
    void Wrong ( const char *what1, const char *what2, const char *what3 );
    void warning ( const char *text );
    void CannotOpen( const char *what );
    void testEnd ();
    void skipspaces();  // it skips spaces and increments line
    void skipSpacesAndComments(); // skips spaces and line(s) with comments
    int eoln_skipspaces();
      // it returns 1 if there is eoln
    int eof_skipspaces();
      // it returns 1 if there is eof

    void LabelRequired ( char *buffer, const char *label );

    lfstream & operator >> ( int & );
    lfstream & operator >> ( unsigned int & );
    lfstream & operator >> ( long & );
    lfstream & operator >> ( unsigned long & );
    lfstream & operator >> ( float & );
    lfstream & operator >> ( double & );

    char readYesNo ( char *buffer );
      // reads 0,n,N or 1,y,Y and returns char 0 or 1

    double degsecrad (char *S);
      // reads deg, sec or rad; S is a buffer
    double length2m (char *S);
      // reads length unit, converts it to meter; S is a buffer

/*
    lfstream & operator >> ( Complex & );
    #ifdef __TURBOC__
    friend lfstream _FAR & _Cdecl _FARFUNC operator >> (lfstream _FAR &, Complex _FAR &);
    #else
    lfstream & operator >> ( lfstream &, Complex & );
    #endif
*/

#ifndef GNU_LIBSTD_3
    lfstream & operator >> ( char & );
    lfstream & operator >> ( unsigned char & );
#else
    lfstream & operator >> ( char & );
    lfstream & operator >> ( signed char & );
    lfstream & operator >> ( unsigned char & );
#endif
    lfstream & operator >> ( char * );

    int readStringWithSpaces ( char *S, int max_len );
	/* 
	Reads a string, which can be enclosed in quotes so that it may contain 
	spaces. Quotes inside the string can be included via \"
	Usage: readStringWithSpaces( S, sizeof(S) );
	Returns: 0 if read OK
		 1 if string exceeds the available length
		 2 unpaired quotes or eof
	Usage: for reading a filename, which can contains spaces or it even can 
	be a command (input from a pipe).
	*/


  private:
    void lfexit ( int i );

  };


#endif
